import {CachedPrimeMath, PrimeMath} from "./PrimeMath";


export class Display {
    private readonly canvas: HTMLCanvasElement;

    private _painter: Painter | undefined = undefined;
    public set painter(painter: Painter) {
        this._painter = painter;
        this.reloadPainter();
    }

    private isDrawing: boolean = false;


    constructor(canvas: HTMLCanvasElement) {
        this.canvas = canvas;
        this.reloadPainter();
    }


    public startDrawLoop(): void {
        if (this.isDrawing) return;

        const cb = () => {
            if (!this.isDrawing) return;

            this._painter?.draw(this.canvas);
            window.requestAnimationFrame(cb);
        };

        this.isDrawing = true;
        window.requestAnimationFrame(cb);
    }

    public stopDrawLoop(): void {
        this.isDrawing = false;
    }


    public reloadPainter(): void {
        this.canvas.width = this._painter?.getPixelWidth() ?? 1;
        this.canvas.height = this._painter?.getPixelHeight() ?? 1;
    }
}


export interface Painter {
    getPixelWidth(): number;

    getPixelHeight(): number;

    draw(canvas: HTMLCanvasElement): void;
}

export class GridPainter implements Painter {
    private readonly primeMath: PrimeMath = CachedPrimeMath.getInstance();

    public scale: number = 15;
    public cols: number = 100;
    public max: number = 10_001;

    public get rows(): number {
        return Math.ceil(this.max / this.cols);
    }


    getPixelWidth(): number {
        return this.scale * this.cols;
    }

    getPixelHeight(): number {
        return this.scale * this.rows;
    }


    draw(canvas: HTMLCanvasElement): void {
        const ctx = canvas.getContext("2d")!;

        this.clearCanvas(ctx);
        this.drawGrid(ctx);
        this.drawPrimes(ctx);
    }

    private clearCanvas(ctx: CanvasRenderingContext2D): void {
        ctx.save();
        ctx.fillStyle = "#ffffff";
        ctx.fillRect(0, 0, this.getPixelWidth(), this.getPixelHeight());
        ctx.restore();
    }

    private drawGrid(ctx: CanvasRenderingContext2D): void {
        // TODO: Don't draw grid for non-filled rows
        ctx.save();
        ctx.beginPath();
        ctx.strokeStyle = "#7b7b7b";
        for (let x = 0; x <= this.cols; x++) {
            ctx.moveTo(x * this.scale, 0);
            ctx.lineTo(x * this.scale, this.rows * this.scale);
        }
        for (let y = 0; y <= this.rows; y++) {
            ctx.moveTo(0, y * this.scale);
            ctx.lineTo(this.cols * this.scale, y * this.scale);
        }
        ctx.stroke();
        ctx.restore();
    }

    private drawPrimes(ctx: CanvasRenderingContext2D): void {
        ctx.save();
        ctx.fillStyle = "#ff0000";
        for (let i = 1; i <= this.max; i++) {
            if (this.primeMath.isPrime(i))
                ctx.fillRect(
                    ((i - 1) % this.cols) * this.scale,
                    Math.floor((i - 1) / this.cols) * this.scale,
                    this.scale,
                    this.scale
                );
        }
        ctx.restore();
    }
}
